"""
Easy-Job worker based on python multithreading and Queue
This backend doesn't have any configuration option

"""
from easy_job.workers.common import call_with_retry, empty_callback
from . import BaseInitializer, BaseRunner
import logging
from queue import Empty
import os
import time
from easy_job.workers.mixins import LoggerMixin, StoreResultMixin

__author__ = 'Apollo'


class MTQueueWorker(StoreResultMixin):
    def __init__(self, queue, *args, **options):
        self.queue = queue
        super(MTQueueWorker, self).__init__(*args, **options)

    def callback(self, body):
        from django.utils.module_loading import import_string
        try:
            function_dot_path, parameters = body['function'], body['parameters']
            args, kwargs = parameters['args'], parameters['kwargs']
            retry_policy = parameters.get('retry_policy', None)
            function = import_string(function_dot_path)
            if callable(function):
                before = time.time()
                if retry_policy is not None:
                    result = call_with_retry(function, args, kwargs, retry_policy)
                else:
                    result = function(*args, **kwargs)
                time_to_complete = time.time() - before
                callback_descriptor = parameters.get('callback', {})
                try:
                    callback_function = import_string(callback_descriptor['function'])
                except:
                    callback_function = empty_callback
                callback_args = callback_descriptor.get('args', tuple())
                callback_kwargs = callback_descriptor.get('kwargs', {})
                callback_kwargs.update(
                    {
                        'result': result,
                        'logger': self._logger
                    }
                )
                self.store_result(function_dot_path, callback_function(*callback_args, **callback_kwargs))
            else:
                raise ValueError("{} is not callable".format(function_dot_path))
            self.log(logging.DEBUG, "{} finished in {} seconds ".format(function_dot_path, time_to_complete))
        except Exception as exp:
            self.log(logging.ERROR, body, exception=exp, include_traceback=True)

    def run(self):
        self.log(logging.DEBUG, "Running the MTQueue worker: {}".format(os.getpid()))
        while True:
            try:
                self.callback(self.queue.get(timeout=2))
            except Empty:
                pass


class MTQueueInitializer(BaseInitializer):
    def start(self, no_runner=False):
        from threading import Thread
        from queue import Queue
        queue = Queue()
        logging.getLogger(self.logger).log(logging.DEBUG, "Starting {} MTQueue workers...".format(self.count))
        if not no_runner:
            for process_index in range(self.count):
                worker_instance = MTQueueWorker(
                    result_backend=self.result_backend,
                    queue=queue,
                    logger=self.logger,
                    **self.options
                )
                t = Thread(
                    target=lambda x: x.run(),
                    args=(worker_instance,)
                )
                t.daemon = True
                t.start()
        return MTQueueRunner(queue=queue, logger=self.logger)


class MTQueueRunner(LoggerMixin, BaseRunner):
    def __init__(self, queue, *args, **kwargs):
        self.queue = queue
        self.log(logging.DEBUG, "MTQueue Runner is ready...")
        super(MTQueueRunner, self).__init__(*args, **kwargs)

    def run(self, function, args=None, kwargs=None, retry_policy=None, callback=None):
        self.queue.put({
            "function": function,
            "parameters": {
                "args": args or tuple(),
                "kwargs": kwargs or {},
                "retry_policy": retry_policy,
                "callback": callback or {},
            }
        })
        self.log(logging.DEBUG, "Task received : {} , Queue size : {}".format(function, self.queue.qsize()))



"""
Result backend for Easy-Job which do nothing
there is no available option
"""
from . import BaseResultBackend
import logging

__author__ = 'Apollo'


class DummyBackend(BaseResultBackend):
    def store(self, task_id, result, *args, **kwargs):
        pass



"""
Easy-Job worker based on python multiprocessing Queue
available options to use
    * process_name_template: a template to create process names , you can use {index} in naming processes
        default: MPQueueProcess_{index}

"""
from easy_job.workers.common import call_with_retry
from . import BaseInitializer, BaseRunner
import logging
from queue import Empty
import os
import time
from easy_job.workers.mixins import LoggerMixin, StoreResultMixin

__author__ = 'Apollo'


def worker(worker_instance):
    worker_instance.run()


class MPQueueWorker(StoreResultMixin):
    def __init__(self, queue, *args, **options):
        self.queue = queue
        super(MPQueueWorker, self).__init__(*args, **options)

    def callback(self, body):
        from django.utils.module_loading import import_string
        try:
            function_dot_path, parameters = body['function'], body['parameters']
            args, kwargs = parameters['args'], parameters['kwargs']
            retry_policy = parameters.get('retry_policy', None)
            callback = parameters.get('callback', None)
            function = import_string(function_dot_path)
            if callable(function):
                before = time.time()
                if retry_policy is not None:
                    result = call_with_retry(function, args, kwargs, retry_policy)
                else:
                    result = function(*args, **kwargs)
                time_to_complete = time.time() - before
                if callback is None:
                    self.store_result(function_dot_path, result)
                else:
                    callback_function = import_string(callback['function'])
                    callback_args = callback.get('args', tuple())
                    callback_kwargs = callback.get('kwargs', {})
                    callback_kwargs.update(
                        {
                            'result': result,
                            'logger': self._logger
                        }
                    )
                    self.store_result(function_dot_path, callback_function(*callback_args, **callback_kwargs))
            else:
                raise ValueError("{} is not callable".format(function_dot_path))
            self.log(logging.DEBUG, "{} finished in {} seconds ".format(function_dot_path, time_to_complete))
        except Exception as exp:
            self.log(logging.ERROR, body, exp, include_traceback=True)

    def run(self):
        self.log(logging.DEBUG, "Running the MPQueue worker: {}".format(os.getpid()))
        while True:
            try:
                self.callback(self.queue.get(timeout=2))
            except Empty:
                pass


class MPQueueInitializer(BaseInitializer):
    def start(self, no_runner=False):
        from multiprocessing import Process, Queue
        queue = Queue()
        logging.getLogger(self.logger).log(logging.DEBUG, "Starting {} MPQueue workers...".format(self.count))
        if not no_runner:
            for process_index in range(self.count):
                process_name = self.options.pop('process_name_template',
                                                "MPQueueProcess_{index}").format(index=process_index)
                worker_instance = MPQueueWorker(
                    result_backend=self.result_backend,
                    queue=queue,
                    logger=self.logger,
                    **self.options
                )
                p = Process(
                    name=process_name,
                    target=worker,
                    args=(worker_instance,)
                )
                p.daemon = True
                p.start()
        return MPQueueRunner(queue=queue, logger=self.logger)


class MPQueueRunner(LoggerMixin, BaseRunner):
    def __init__(self, queue, *args, **kwargs):
        self.queue = queue
        self.log(logging.DEBUG, "MPQueue Runner is ready...")
        super(MPQueueRunner, self).__init__(*args, **kwargs)

    def run(self, function, args=None, kwargs=None, retry_policy=None, callback=None):
        self.queue.put({
            "function": function,
            "parameters": {
                "args": args or tuple(),
                "kwargs": kwargs or {},
                "retry_policy": retry_policy,
                "callback": callback,
            }
        })
        self.log(logging.DEBUG, "Task received : {} , Queue size : {}".format(function, self.queue.qsize()))



import logging
from django.utils.module_loading import import_string

__author__ = 'Apollo'


class LoggerMixin(object):
    _logger = None

    def __init__(self, logger, ):
        self.logger = logger
        self._logger = logging.getLogger(logger)

    def log(self, level, message, **kwargs):
        if hasattr(self, 'logger'):
            import traceback
            if self._logger is None:
                self._logger = logging.getLogger(self.logger)
            self._logger.log(
                level,
                "{} : {}\n{}".format(
                    message,
                    kwargs.get("exception") or "",
                    traceback.format_exc() if kwargs.get('include_traceback', False) else ""
                )
            )


class StoreResultMixin(LoggerMixin):
    def __init__(self, result_backend=None, *args, **options):
        try:
            if result_backend is None:
                raise ValueError("if you don't want a result_backend use dummy result backend")
                # noinspection PyPep8Naming
            ResultBackend = import_string(result_backend['result_backend_class'])
            self.result_backend = ResultBackend(**result_backend.get('options', {}))
        except KeyError:
            raise KeyError("You have to provide result_backend_class for result backend")
        super(StoreResultMixin, self).__init__(*args, **options)

    def store_result(self, task_id, result):
        try:
            self.result_backend.store(task_id, result)
        except Exception as exp:
            self.log(logging.ERROR, "problem while logging result", exp, include_traceback=True)



from retrying import Retrying

__author__ = "Apollo"


def call_with_retry(function, args: tuple, kwargs: dict, retry_policy: dict):
    return Retrying(**retry_policy).call(function, *args, **kwargs)


def empty_callback(result, logger, *args, **kwargs):
    return result



"""
Result backend for Easy-Job which use python logging
available options to use
    * log_level: level of logging could be logging.DEBUG logging.ERROR or any other log level defined in logging module
        default: logging.DEBUG
    * logger: name of the logger to use, be cautious , the logger must already defined
        default: no default value available, you must specify either logger or logger_instance
    * logger_instance: for testing purposes you can provide the logger object itself
        default: None
    message_template: string template to log function and result , you can use {task_id} and {result} in your template
        default: "{task_id} -> {result}"
"""
from . import BaseResultBackend
import logging

__author__ = 'Apollo'


class LogResultBackend(BaseResultBackend):
    def __init__(self, log_level=logging.DEBUG, logger=None, logger_instance=None, message_template=None):
        if logger_instance is not None:
            self.logger = logger_instance
        elif logger is not None:
            self.logger = logging.getLogger(logger)
        else:
            raise ValueError("No logger or logger_instance specified")
        self.log_level = log_level
        self.message_template = message_template or "{task_id} -> {result}"

    def store(self, task_id, result, *args, **kwargs):
        self.logger.log(
            level=self.log_level,
            msg=self.message_template.format(
                task_id=task_id,
                result=result
            )
        )





"""
Easy-Job worker based on RabbitMQ
available options to use
    * queue_name: name of the underlying rabbitmq queue
        default: no default value available you have to specify queue name or ValueError will be raise
    * serialization_method: method of serializing messages before putting them in the queue , can be json or pickle
        default: no default value available you have to specify serialization_method
    * rabbitmq_configs: a dictionary of rabbitmq related configurations consist of :
        * connection_pool_configs : which is another nested dictionary consist of :
            * max_size: Maximum number of connections to keep queued , default is 10
            * max_overflow: Maximum number of connections to create above `max_size`, default is 10
            * timeout: Default number of seconds to wait for a connections to available, default is 30
            * recycle: Lifetime of a connection (since creation) in seconds or None for no
                       recycling. Expired connections are closed on acquire , default is None
            * stale: Threshold at which inactive (since release) connections are
                     considered stale in seconds or None for no staleness. Stale
                     connections are closed on acquire , default is None
        * connection_parameters: which is another nested dictionary consist of parameters to send
                                to pika.ConnectionParameters

"""
from . import BaseInitializer, BaseRunner
import logging
import os
from django.utils.functional import SimpleLazyObject
import pika
import pika_pool
import time
from easy_job.workers.mixins import StoreResultMixin, LoggerMixin
from .common import call_with_retry
__author__ = 'Apollo'


def worker(worker_instance):
    worker_instance.run()


class RabbitMQWorker(StoreResultMixin):
    def __init__(self, queue_name, rabbitmq_configs, serialization_method, *args, **kwargs):
        self.connection_params = rabbitmq_configs.get('connection_parameters', {})
        self.serialization_method = serialization_method
        self.queue_name = queue_name
        super(RabbitMQWorker, self).__init__(*args, **kwargs)

    def deserialize(self, body):
        if self.serialization_method == "json":
            import json
            return json.loads(body.decode('utf8'))
        elif self.serialization_method == "pickle":
            import pickle
            return pickle.loads(body)
        else:
            raise ValueError("unknown serialization_method")

    def callback(self, ch, method, properties, body):
        from django.utils.module_loading import import_string
        try:
            body = self.deserialize(body)
            function_dot_path, parameters = body['function'], body['parameters']
            args, kwargs = parameters['args'], parameters['kwargs']
            retry_policy = parameters.get('retry_policy', None)
            callback = parameters.get('callback', None)
            function = import_string(function_dot_path)
            if callable(function):
                before = time.time()
                if retry_policy is not None:
                    result = call_with_retry(function, args, kwargs, retry_policy)
                else:
                    result = function(*args, **kwargs)
                time_to_complete = time.time() - before
                if callback is None:
                    self.store_result(function_dot_path, result)
                else:
                    callback_function = import_string(callback['function'])
                    callback_args = callback.get('args', tuple())
                    callback_kwargs = callback.get('kwargs', {})
                    callback_kwargs.update(
                        {
                            'result': result,
                            'logger': self._logger
                        }
                    )
                    self.store_result(function_dot_path, callback_function(*callback_args, **callback_kwargs))
            else:
                raise ValueError("{} is not callable".format(function_dot_path))
            self.log(logging.DEBUG, "{} finished in {} seconds ".format(function_dot_path, time_to_complete))
        except Exception as exp:
            self.log(logging.ERROR, body, exception=exp, include_traceback=True)
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def run(self):
        self.log(logging.DEBUG, "Running the RabbitMQ worker: {}".format(os.getpid()))
        with pika.BlockingConnection(pika.ConnectionParameters(**self.connection_params)) as connection:
            channel = connection.channel()
            channel.queue_declare(queue=self.queue_name, durable=True)
            channel.basic_qos(prefetch_count=1)
            channel.basic_consume(self.callback, queue=self.queue_name)
            channel.start_consuming()


class RabbitMQInitializer(BaseInitializer):
    def start(self, no_runner=False):
        from multiprocessing import Process

        serialization_method = self.options['serialization_method']
        if serialization_method == "json":
            import json as serialization_method
        elif serialization_method == "pickle":
            import pickle as serialization_method
        else:
            raise ValueError("unknown serialization method")
        logging.getLogger(self.logger).log(logging.DEBUG, "Starting {} RabbitMQ workers...".format(self.count))

        if not no_runner:

            for process_index in range(self.count):
                # here the multiprocessing logic
                # create a rabbitmq worker to send for worker function
                if 'queue_name' not in self.options or self.options['queue_name'] == "":
                    raise ValueError("you must provide queue_name in worker options")
                worker_instance = RabbitMQWorker(
                    result_backend=self.result_backend,
                    logger=self.logger,
                    **self.options
                )

                p = Process(target=worker, args=(worker_instance,))
                p.daemon = True
                p.start()
        return RabbitMQRunner(queue_name=self.options['queue_name'],
                              serializer=serialization_method.dumps,
                              rabbitmq_configs=self.options.pop("rabbitmq_configs"),
                              logger=self.logger)


class RabbitMQRunner(LoggerMixin, BaseRunner):
    def __init__(self, queue_name, serializer, rabbitmq_configs, *args, **kwargs):
        self.queue_name = queue_name
        self.serialize = serializer
        super(RabbitMQRunner, self).__init__(*args, **kwargs)
        self.log(logging.DEBUG, "RabbitMQ Runner is ready...")

        def _create_pool():
            connection_pool_configs = rabbitmq_configs.get('connection_pool_configs', {})

            def create_connection():
                self.log(logging.DEBUG, "Creating new rabbitmq connection")
                con_params = pika.ConnectionParameters(**rabbitmq_configs.get('connection_parameters', {}))
                return pika.BlockingConnection(con_params)

            return pika_pool.QueuedPool(
                create=create_connection,
                **connection_pool_configs
            )

        self._pool = SimpleLazyObject(_create_pool)

    def run(self, function, args=None, kwargs=None, retry_policy=None, callback=None):
        with self._pool.acquire() as cxn:
            cxn.channel.basic_publish(
                body=self.serialize(
                    {
                        "function": function,
                        "parameters": {
                            "args": args or tuple(),
                            "kwargs": kwargs or {},
                            "retry_policy": retry_policy,
                            "callback": callback
                        }
                    }
                ),
                exchange='',
                routing_key=self.queue_name,
                properties=pika.BasicProperties(
                    delivery_mode=2,
                )
            )
            self.log(logging.DEBUG, "Task received : {}".format(function))



import logging

__author__ = 'Apollo'


class BaseInitializer(object):
    def __init__(self, count, logger, options, result_backend=None, *args, **kwargs):
        self.count = count
        self.result_backend = result_backend
        self.logger = logger
        self.options = options

    def start(self, no_runner=False):
        """
        start workers and return a runner instance

        :return: a runner instance ro run
        :rtype: BaseRunner
        """
        raise NotImplementedError()

    def log_error(self, message, exception=None, include_traceback=True):
        import traceback
        logging.getLogger(self.logger).error(
            "{} : {}\n{}".format(
                message,
                exception or "",
                traceback.format_exc() if include_traceback else ""
            )
        )


class BaseRunner(object):
    def run(self, function, args, kwargs):
        pass
